_resource_classes = \
    {
    ### XPROTO ###
      "WINDOW"
    , "PIXMAP"
    , "CURSOR"
    , "FONT"
    , "GCONTEXT"
    , "COLORMAP"
    , "ATOM"
    , "DRAWABLE"
    , "FONTABLE"
    ### XPROTO ###

    ### DAMAGE ###
    , "DAMAGE"
    ### DAMAGE ###

    ### GLX, RECORD, XVMC ###
    , "CONTEXT"
    ### GLX, RECORD, XVMC ###

    ### GLX ###
    # , "PIXMAP" # already in XPROTO
    # , "CONTEXT"
    , "PBUFFER"
    # , "WINDOW" # already in XPROTO
    , "FBCONFIG"
    ### GLX ###

    ### PRESENT ###
    , "EVENT"
    ### PRESENT ###

    ### RANDR ###
    , "MODE"
    , "CRTC"
    , "OUTPUT"
    , "PROVIDER"
    ### RANDR ###

    ### RECORD ###
    # , "CONTEXT"
    ### RECORD ###

    ### RENDER ###
    , "GLYPHSET"
    , "PICTURE"
    , "PICTFORMAT"
    ### RENDER ###

    ### SHM ###
    , "SEG"
    ### SHM ###

    ### SYNC ###
    , "ALARM"
    , "COUNTER"
    , "FENCE"
    ### SYNC ###

    ### XFIXES ###
    , "REGION"
    , "BARRIER"
    ### XFIXES ###

    ### XPRINT ###
    , "PCONTEXT"
    ### XPRINT ###

    ### XVMC ###
    # , "CONTEXT"
    , "SURFACE"
    , "SUBPICTURE"
    ### XVMC ###

    ### XV ###
    , "PORT"
    , "ENCODING"
    ### XV ###
    }



import re # compile

_reserved_keywords = {'class' : '_class',
                      'new'   : '_new',
                      'delete': '_delete',
                      'default' : '_default',
                      'private' : '_private',
                      'explicit': '_explicit'}

def get_namespace(namespace):
    if namespace.is_ext:
        return get_ext_name(namespace.ext_name)
    else:
        return "x"

def get_ext_name(string):
    return _ext(string)

_cname_re = re.compile('([A-Z0-9][a-z]+|[A-Z0-9]+(?![a-z])|[a-z]+)')
_cname_special_cases = {'DECnet':'decnet'}

def _n_item(string, parts=False):
    '''
    Does C-name conversion on a single string fragment.
    The resulting string is a valid C-name
    Uses a regexp with some hard-coded special cases.
    '''
    if string in _cname_special_cases:
        return _cname_special_cases[string]
    else:
        split = _cname_re.finditer(string)
        name_parts = [match.group(0) for match in split]
        if parts:
            return name_parts
        else:
            return '_'.join(name_parts)

_extension_special_cases = ['XPrint', 'XCMisc', 'BigRequests']

def _ext(string):
    '''
    Does C-name conversion on an extension name.
    Has some additional special cases on top of _n_item.
    '''
    if string in _extension_special_cases:
        return _n_item(string).lower()
    else:
        return string.lower()

def _n(list, namespace):
    '''
    Does C-name conversion on a tuple of strings.
    Different behavior depending on length of tuple, extension/not extension, etc.
    Basically C-name converts the individual pieces, then joins with underscores.
    '''
    if len(list) == 1:
        parts = list
    elif len(list) == 2:
        parts = [list[0], _n_item(list[1])]
    elif namespace.is_ext:
        parts = [list[0], _ext(list[1])] + [_n_item(i) for i in list[2:]]
    else:
        parts = [list[0]] + [_n_item(i) for i in list[1:]]
    return '_'.join(parts).lower()



from utils import \
        get_namespace, \
        get_ext_name, \
        _n_item, \
        _ext

class ExtensionClass(object):
    def __init__(self, namespace):
        self.namespace = namespace

    def make_class(self):
        # if not self.namespace.is_ext:
        #     return ""
        # else:
        ns = get_namespace(self.namespace)
        if self.namespace.is_ext:
            base = "\n  : public xpp::generic::extension<extension, &xcb_%s_id>\n" % ns
            ctor = "    using base = xpp::generic::extension<extension, &xcb_%s_id>;\n" % ns + \
                   "    using base::base;\n"
        else:
            base = " "
            ctor = ""

        return \
'''\
template<typename Derived, typename Connection>
class interface;

namespace event { template<typename Connection> class dispatcher; }
namespace error { class dispatcher; }

class extension%s{
  public:
%s\
    template<typename Derived, typename Connection>
    using interface = xpp::%s::interface<Derived, Connection>;
    template<typename Connection>
    using event_dispatcher = xpp::%s::event::dispatcher<Connection>;
    using error_dispatcher = xpp::%s::error::dispatcher;
};\
''' % (base,
       ctor,
       ns, # typedef xpp::interface::%s interface;
       ns, # typedef xpp::event::dispatcher::%s dispatcher;
       ns) # typedef xpp::error::dispatcher::%s dispatcher;



# vim: set ts=4 sws=4 sw=4:

import sys # stderr

_templates = {}

_templates['initializer'] = \
'''\
typedef typename value_type<%s, ! std::is_pointer<%s>::value>::type
          vector_type;
std::vector<vector_type> %s =
  { value_iterator<%s>(%s), value_iterator<%s>(%s) };
'''

def _initializer(iter_type, c_name, iter_begin, iter_end):
    return _templates['initializer'] % \
            ( iter_type
            , iter_type
            , c_name
            , iter_type
            , iter_begin
            , iter_type
            , iter_end
            )

class ParameterList(object):
    def __init__(self):
        self.want_wrap = False
        self.has_defaults = False
        self.parameter = []
        self.wrap_calls = []
        self.wrap_protos = []
        self.iter_calls = []
        self.iter_2nd_lvl_calls = []
        self.iter_protos = []
        self.templates = []
        self.iterator_templates = []
        self.initializer = []

    def add(self, param):
        self.has_defaults = param.default != None
        self.parameter.append(param)

    def comma(self):
        return "" if len(self.parameter) == 0 else ", "

    def is_reordered(self):
        tmp = sorted(self.parameter, key=lambda p: p.default or '')
        return tmp != self.parameter

    def calls(self, sort, params=None):
        ps = self.parameter if params == None else params
        if sort:
            tmp = sorted(ps, key=lambda p: p.default or '')
            ps = tmp
        calls = [p.call() for p in ps]
        return ", ".join(calls)

    def protos(self, sort, defaults, params=None):
        if defaults: sort = True
        ps = self.parameter if params == None else params
        if sort:
            tmp = sorted(ps, key=lambda p: p.default or '')
            ps = tmp
        protos = [p.proto(defaults) for p in ps]
        return ", ".join(protos)

    def iterator_initializers(self):
        return self.initializer

    def make_wrapped(self):
        self.wrap_calls = []
        self.wrap_protos = []
        self.iter_calls = []
        self.iter_2nd_lvl_calls = []
        self.iter_protos = []
        self.initializer = []
        self.templates = []
        self.iterator_templates = []

        lenfields = {}
        # if a parameter is removed, take reduced parameter size into account
        adjust = 0
        for index, param in enumerate(self.parameter):
            prev = index - adjust - 1

            if param.field.type.is_list:
                name = param.field.type.expr.lenfield_name
                if name in lenfields:
                    lenfields[name].append(param.c_name)
                else:
                    lenfields[name] = [ param.c_name ]

                # sys.stderr.write("list: %s %s\n\n"
                #         % ( param.field.type.expr.lenfield_type
                #           , param.field.type.expr.lenfield_name
                #           ))

            # SetGamma: takes 1 size, but 3 value lists
            # if param.field.type.is_list and prev >= 0:
            if (param.is_const and param.is_pointer
                    and prev >= 0
                    and self.parameter[prev].c_name == param.c_name + "_len"):

                adjust = adjust + 1
                self.want_wrap = True
                self.wrap_calls.pop(prev)
                self.wrap_protos.pop(prev)
                self.iter_calls.pop(prev)
                self.iter_2nd_lvl_calls.pop(prev)
                self.iter_protos.pop(prev)

                prev_type = self.parameter[prev].c_type
                if param.c_type == 'char':

                    def append_proto_string(list):
                        list.append(Parameter(None, \
                            c_type='const std::string &',
                            c_name=param.c_name))

                    def append_call_string(list):
                        list.append(Parameter(None, \
                            c_name="static_cast<" + prev_type + ">(" \
                            + param.c_name + '.length())'))

                        list.append(Parameter(None, \
                            c_name=param.c_name + '.c_str()'))

                    append_proto_string(self.wrap_protos)
                    append_proto_string(self.iter_protos)
                    append_call_string(self.wrap_calls)
                    append_call_string(self.iter_calls)
                    append_call_string(self.iter_2nd_lvl_calls)

                else:
                    param_type = param.c_type
                    if param_type == "void":
                        param_type = "Type_" + str(index)
                        self.templates.append(param_type)

                    prev_type = self.parameter[prev].c_type

                    ### std::vector
                    self.wrap_protos.append(Parameter(None, \
                        c_type='const std::vector<' + param_type + '> &',
                        c_name=param.c_name))

                    self.wrap_calls.append(Parameter(None, \
                      c_name="static_cast<" + prev_type + ">(" \
                      + param.c_name + '.size())'))

                    self.wrap_calls.append(Parameter(None, \
                        c_name=param.c_name + '.data()'))

                    ### Iterator
                    iter_type = param.c_name.capitalize() + "_Iterator"
                    iter_begin = param.c_name + "_begin"
                    iter_end = param.c_name + "_end"

                    if len(self.templates) > 0:
                        self.templates[-1] += " = typename " + iter_type + "::value_type"
                    self.iterator_templates.append(iter_type)

                    self.iter_protos.append(Parameter(None, \
                            c_type=iter_type,
                            c_name=iter_begin))

                    self.iter_protos.append(Parameter(None, \
                            c_type=iter_type,
                            c_name=iter_end))

                    self.iter_calls.append(Parameter(None, \
                            c_name="static_cast<" + prev_type + ">(" \
                            + param.c_name + '.size())'))

                    self.iter_calls.append(Parameter(None, \
                            c_name='const_cast<const vector_type *>(' \
                            + param.c_name + '.data())'))

                    self.iter_2nd_lvl_calls.append(Parameter(None, \
                            c_name=iter_begin))

                    self.iter_2nd_lvl_calls.append(Parameter(None, \
                            c_name=iter_end))

#                     vector_type = \
#                     '''\
# typename value_type<%s,
#                   ! std::is_pointer<%s>::value
#                  >::type\
#                     ''' % (iter_type, iter_type)

                    # self.initializer.append( \
                    #         "std::vector<%s> %s = { value_iterator<%s>(%s), \
                    #         value_iterator<%s>(%s) };" \
                    #         % (vector_type, param.c_name,
                    #             iter_type, iter_begin,
                    #             iter_type, iter_end))

                    self.initializer.append(
                            _initializer(iter_type, param.c_name, iter_begin, iter_end))

            else:
                self.wrap_calls.append(param)
                self.wrap_protos.append(param)
                self.iter_calls.append(param)
                self.iter_2nd_lvl_calls.append(param)
                self.iter_protos.append(param)

        # end: for index, param in enumerate(self.parameter):

        for k, v in list(lenfields.items()):
            if len(v) > 1:
                sys.stderr.write("list: %s, %s\n" % (k, v))


    def wrapped_calls(self, sort):
        return self.calls(sort, params=self.wrap_calls)

    def wrapped_protos(self, sort, defaults):
        return self.protos(sort, defaults, params=self.wrap_protos)

    def iterator_calls(self, sort):
        return self.calls(sort, params=self.iter_calls)

    def iterator_2nd_lvl_calls(self, sort):
        return self.calls(sort, params=self.iter_2nd_lvl_calls)

    def iterator_protos(self, sort, defaults):
        return self.protos(sort, defaults, params=self.iter_protos)



_default_parameter_values = \
    { "xcb_timestamp_t" : "XCB_TIME_CURRENT_TIME" }

class Parameter(object):
    def __init__(self, field, c_type="", c_name="", verbose=False):
        self.field = field
        if field != None:
            self.c_type = field.c_field_type
            self.c_name = field.c_field_name
            self.is_const = field.c_field_const_type == "const " + field.c_field_type
            self.is_pointer = field.c_pointer != " "
            # self.serialize = field.type.need_serialize
            self.default = _default_parameter_values.get(self.c_type)
            self.with_default = True
            if verbose:
                sys.stderr.write("c_type: %s; c_name: %s; default: %s\n" \
                      % (self.c_type, self.c_name, self.default))

        else:
            self.c_type = c_type
            self.c_name = c_name
            self.is_const = False
            self.is_pointer = False
            # self.serialize = field.type.need_serialize
            self.default = _default_parameter_values.get(self.c_type)
            self.with_default = True

    def call(self):
        return self.c_name

    def proto(self, with_default):
        c_type = ("const " if self.is_const else "") \
             + self.c_type \
             + (" *" if self.is_pointer else "")
        param = " = " + self.default if with_default and self.default != None else ""
        return c_type + " " + self.c_name + param



# vim: set ts=4 sws=4 sw=4:

import sys # stderr
import copy # deepcopy

from utils import \
        get_namespace, \
        get_ext_name, \
        _n_item, \
        _ext

class ObjectClass(object):
    def __init__(self, name):
        self.name = name
        self.requests = []

    def add(self, request):
        if (len(request.parameter_list.parameter) > 0
                and request.parameter_list.parameter[0].c_type == self.c_name):
            request_copy = copy.deepcopy(request)
            request_copy.parameter_list.parameter.pop(0)
            request_copy.make_wrapped()
            self.requests.append(request_copy)

    def set_namespace(self, namespace):
        self.namespace = namespace
        name = (get_namespace(namespace) + "_") if namespace.is_ext else ""
        self.c_name = "xcb_%s_t" % (name + self.name.lower())

    def make_inline(self):
        ns = get_namespace(self.namespace)
        name = self.name.lower()
        c_name = self.c_name
        methods = ""

        for request in self.requests:
            methods += request.make_object_class_inline(False, self.name.lower()) + "\n\n"

        if methods == "":
            return ""
        else:
            return \
"""\
template<typename Derived, typename Connection>
class %s
{
  protected:
    Connection
    connection(void) const
    {
      return static_cast<const Derived *>(this)->connection();
    }

    const %s &
    resource(void) const
    {
      return static_cast<const Derived *>(this)->resource();
    }

  public:
    virtual ~%s(void) {}

%s
}; // class %s
""" % (name,   # class %s
       c_name, # %s resource(void) { ... }
       name, # virtual ~%s(void)
       methods,
       name) # }; // class %s



from utils import _n, _ext, _n_item, get_namespace
from resource_classes import _resource_classes

_templates = {}

_templates['reply_class'] = \
'''\
namespace reply {

namespace detail {

template<typename Connection,
         typename Check,
         typename CookieFunction>
class %s
  : public xpp::generic::reply<%s<Connection, Check, CookieFunction>,
                               Connection,
                               Check,
                               SIGNATURE(%s_reply),
                               CookieFunction>
{
  public:
    typedef xpp::generic::reply<%s<Connection, Check, CookieFunction>,
                                Connection,
                                Check,
                                SIGNATURE(%s_reply),
                                CookieFunction>
                                  base;

    template<typename C, typename ... Parameter>
    %s(C && c, Parameter && ... parameter)
      : base(std::forward<C>(c), std::forward<Parameter>(parameter) ...)
    {}

%s\
%s\
}; // class %s

} // namespace detail

namespace checked {
template<typename Connection>
using %s = detail::%s<
    Connection, xpp::generic::checked_tag,
    SIGNATURE(%s)>;
} // namespace checked

namespace unchecked {
template<typename Connection>
using %s = detail::%s<
    Connection, xpp::generic::unchecked_tag,
    SIGNATURE(%s_unchecked)>;
} // namespace unchecked

} // namespace reply
'''

def _reply_class(name, c_name, ns, cookie, accessors):
    return _templates['reply_class'] % \
            ( name
            , name # base class
            , c_name # %s_reply
            , name # typedef
            , c_name # %s_reply
            , name # c'tor
            , cookie.make_static_getter()
            , accessors
            , name # // class %s
            , name # checked { using %s =
            , name # checked { detail::%s
            , c_name # checked { SIGNATURE
            , name # unchecked { using %s =
            , name # unchecked { detail::%s
            , c_name # unchecked { SIGNATURE
            )

_templates['reply_member_accessor'] = \
'''\
    template<typename ReturnType = %s, typename ... Parameter>
    ReturnType
    %s(Parameter && ... parameter)
    {
      using make = xpp::generic::factory::make<Connection,
                                               decltype(this->get()->%s),
                                               ReturnType,
                                               Parameter ...>;
      return make()(this->m_c,
                    this->get()->%s,
                    std::forward<Parameter>(parameter) ...);
    }
'''

def _reply_member_accessor(request_name, name, c_type, template_type):
    return _templates['reply_member_accessor'] % \
            ( c_type
            , name
            , name
            , name
            )

class CppReply(object):
    def __init__(self, namespace, name, cookie, reply, accessors, parameter_list):
        self.namespace = namespace
        self.name = name
        self.reply = reply
        self.cookie = cookie
        self.accessors = accessors
        self.parameter_list = parameter_list
        self.request_name = _ext(_n_item(self.name[-1]))
        self.c_name = "xcb" \
            + (("_" + get_namespace(namespace)) if namespace.is_ext else "") \
            + "_" + self.request_name

    def make_accessors(self):
        return "\n".join(["\n%s\n" % a for a in self.accessors])

    def make(self):
        accessors = [self.make_accessors()]
        naccessors = len(self.accessors)

        for field in self.reply.fields:
            if (field.field_type[-1] in _resource_classes
                and not field.type.is_list
                and not field.type.is_container):

                naccessors = naccessors + 1

                name = field.field_name.lower()
                c_type = field.c_field_type
                template_type = field.field_name.capitalize()

                accessors.append(_reply_member_accessor(self.request_name, name, c_type, template_type))

        result = ""
        result += _reply_class(
            self.request_name, self.c_name, get_namespace(self.namespace),
            self.cookie, "\n".join(accessors))
        return result



from utils import \
        get_namespace, \
        get_ext_name, \
        _n_item, \
        _ext, \
        _reserved_keywords

_templates = {}

_templates['error_dispatcher_class'] = \
'''\
namespace error {

class dispatcher
{
  public:
%s\
%s\

    void
    operator()(const std::shared_ptr<xcb_generic_error_t> &%s) const
    {
%s\
    }

%s\
}; // class dispatcher

} // namespace error
'''

def _error_dispatcher_class(typedef, ctors, switch, members, has_errors):
    return _templates['error_dispatcher_class'] % \
        ( typedef
        , ctors
        , " error" if has_errors else ""
        , switch if has_errors else ""
        , members
        )


def error_dispatcher_class(namespace, cpperrors):
    ns = get_namespace(namespace)

    ctor_name = "dispatcher"

    typedef = []
    ctors = []
    members = []
    opcode_switch = "error->error_code"

    typedef = [ "typedef xpp::%s::extension extension;\n" % ns ]

    # >>> if begin <<<
    if namespace.is_ext:
        opcode_switch = "error->error_code - m_first_error"

        members += \
            [ "protected:"
            , "  uint8_t m_first_error;"
            ]

        ctors = \
            [ "%s(uint8_t first_error)" % (ctor_name)
            , "  : m_first_error(first_error)"
            , "{}"
            , ""
            , "%s(const xpp::%s::extension & extension)" % (ctor_name, ns)
            , "  : %s(extension->first_error)" % ctor_name
            , "{}"
            ]

    # >>> if end <<<

    if len(typedef) > 0:
        typedef = "\n".join(["    " + s for s in typedef]) + "\n"
    else:
        typedef = ""

    if len(ctors) > 0:
        ctors = "\n".join([("    " if len(s) > 0 else "") + s for s in ctors]) + "\n"
    else:
        ctors = ""

    if len(members) > 0:
        members = "\n".join(["  " + s for s in members]) + "\n"
    else:
        members = ""

    switch = error_switch_cases(cpperrors, opcode_switch, "error")
    return _error_dispatcher_class(typedef,
                                   ctors,
                                   switch,
                                   members,
                                   len(cpperrors) > 0)

def error_switch_cases(cpperrors, arg_switch, arg_error):
    cases = ""
    errors = cpperrors
    templ = [ "        case %s: // %s"
            , "          throw %s" + "(%s);" % arg_error
            , ""
            , ""
            ]

    cases += "      switch (%s) {\n\n" % arg_switch
    for e in errors:
        cases += "\n".join(templ) % (e.opcode_name, e.opcode, e.scoped_name())
    cases += "      };\n"

    return cases


class CppError(object):
    def __init__(self, error, namespace, name, c_name, opcode, opcode_name):
        self.error = error
        self.namespace = namespace
        self.c_name = c_name
        self.opcode = opcode
        self.opcode_name = opcode_name

        self.names = list(map(str.lower, _n_item(name[-1], True)))
        self.name = "_".join(self.names)

        self.nssopen = ""
        self.nssclose = ""
        self.scope = []
        for name in self.names[0:-1]:
            if name in _reserved_keywords: name += "_"
            self.nssopen += " namespace %s {" % name
            self.nssclose += " }"
            self.scope.append(name)

    def get_name(self):
        return _reserved_keywords.get(self.name, self.name)


    def scoped_name(self):
        ns = get_namespace(self.namespace)
        return "xpp::" + ns + "::error::" + self.get_name()

    def make_class(self):
        ns = get_namespace(self.namespace)
        typedef = []
        members = []

        opcode_accessor = \
            [ "static uint8_t opcode(void)"
            , "{"
            , "  return %s;" % self.opcode_name
            , "}"
            ]

        if self.namespace.is_ext:
            opcode_accessor += \
                [ ""
                , "static uint8_t opcode(uint8_t first_error)"
                , "{"
                , "  return first_error + opcode();"
                , "}"
                , ""
                , "static uint8_t opcode(const xpp::%s::extension & extension)" % ns
                , "{"
                , "  return opcode(extension->first_error);"
                , "}"
                ]

            members = \
                [ "protected:"
                , "  uint8_t m_first_error;"
                ]

        if len(opcode_accessor) > 0:
            opcode_accessor = "\n".join(["    " + s for s in opcode_accessor]) + "\n"
        else:
            opcode_accessor = ""

        if len(members) > 0:
            members = "\n" + "\n".join(["  " + s for s in members]) + "\n"
        else:
            members = ""

        if len(typedef) > 0:
            typedef = "\n".join(["    " + s for s in typedef]) + "\n\n"
        else:
            typedef = ""

        name = self.name
        if self.name in _reserved_keywords: name = self.name + "_"

        return \
'''
namespace error {
class %s
  : public xpp::generic::error<%s,
                               %s>
{
  public:
%s\
    using xpp::generic::error<%s, %s>::error;

    virtual ~%s(void) {}

%s
    static std::string description(void)
    {
      return std::string("%s");
    }
%s\
}; // class %s
} // namespace error
''' % (self.get_name(), # class %s
       self.get_name(), # : public xpp::generic::error<%s,
       self.c_name, # %s>
       typedef,
       self.get_name(), self.c_name, # using xpp::generic::error<%s, %s>::error;
       self.get_name(), # virtual ~%s(void) {}
       opcode_accessor,
       self.opcode_name, # static constexpr const char * opcode_literal
       members,
       self.get_name()) # // class %s



#!/usr/bin/env python3

from pathlib import Path
import sys
import os
import argparse
import subprocess

EXTENSIONS = set('.' + ext for ext in ['c', 'h', 'cpp', 'hpp', 'inl'])


def get_files(dirs):
    """
    Generator which yields all files in the given directories with any of the
    EXTENSIONS.
    """
    for dir in dirs:
        for root, _, files in os.walk(dir):
            for file in files:
                path = Path(os.path.join(root, file))
                if path.suffix in EXTENSIONS:
                    yield path


def main():
    parser = argparse.ArgumentParser(
        description="""
        Run command on all C/C++ source files in the given directories
        """)
    parser.add_argument('--dirs', type=Path, nargs='+',
                        help='Directories to search in')
    parser.add_argument('command', nargs='+',
                        help='Command to which to pass found files')
    args = parser.parse_args()

    all_files = list(str(file) for file in get_files(args.dirs))

    if not all_files:
        print("No files found")
        sys.exit(1)

    result = subprocess.run(args.command + all_files)
    print(f'Formatted {len(all_files)} files')

    if result.returncode != 0:
        sys.exit(result.returncode)


if __name__ == '__main__':
    main()



from resource_classes import _resource_classes

_templates = {}

_templates['iter_fixed'] = \
"""\
xpp::generic::iterator<Connection,
                       %s,
                       SIGNATURE(%s_%s),
                       SIGNATURE(%s_%s_length)>\
"""

_templates['iter_variable'] = \
"""\
xpp::generic::iterator<Connection,
                       %s,
                       SIGNATURE(%s_next),
                       SIGNATURE(%s_sizeof),
                       SIGNATURE(%s_%s_iterator)>\
"""

_templates['list'] = \
"""\
    xpp::generic::list<Connection,
                       %s_reply_t,
                       %s
                      >
    %s(void)
    {
      return xpp::generic::list<Connection,
                                %s_reply_t,
                                %s
                               >(%s);
    }\
"""

_templates['string_accessor'] = \
'''\
    std::string
    %s(void)
    {
      return std::string(%s_%s(this->get().get()),
                         %s_%s_length(this->get().get()));
    }
'''

def _string_accessor(member, c_name):
    return _templates['string_accessor'] % \
            (member, c_name, member, c_name, member)

class Accessor(object):
    def __init__(self, is_fixed=False, is_string=False, is_variable=False, \
                 member="", c_type="", return_type="", iter_name="", c_name=""):

        self.is_fixed = is_fixed
        self.is_string = is_string
        self.is_variable = is_variable

        self.member = member
        self.c_type = c_type
        self.return_type = return_type
        self.iter_name = iter_name
        self.c_name = c_name

        self.object_type = self.c_type.replace("xcb_", "").replace("_t", "").upper()

        if self.c_type == "void":
            self.return_type = "Type"
        elif self.object_type in _resource_classes:
            self.return_type = self.member.capitalize()
        else:
            self.return_type = self.c_type

    def __str__(self):
        if self.is_fixed:
            return self.list(self.iter_fixed())
        elif self.is_variable:
            return self.list(self.iter_variable())
        elif self.is_string:
            return self.string()
        else:
            return ""


    def iter_fixed(self):
        return_type = self.return_type

        return _templates['iter_fixed'] \
                % (return_type,
                   self.c_name, self.member,
                   self.c_name, self.member)


    def iter_variable(self):
        return _templates['iter_variable'] \
                % (self.c_type,
                   self.iter_name,
                   self.iter_name,
                   self.c_name, self.member)


    def list(self, iterator):
        template = "    template<typename Type" if self.c_type == "void" else ""

        # template<typename Children = xcb_window_t>
        if self.object_type in _resource_classes:
            template += ", " if template != "" else "    template<typename "
            template += self.member.capitalize() + " = " + self.c_type

        template += ">\n" if template != "" else ""

        c_tor_params = "this->m_c, this->get()"

        fst_iterator = "\n                       ".join(iterator.split('\n'))
        snd_iterator = "\n                                ".join(iterator.split('\n'))

        return template + _templates['list'] \
                % (self.c_name,
                   fst_iterator,
                   self.member,
                   self.c_name,
                   snd_iterator,
                   c_tor_params)

    def string(self):
        return _string_accessor(self.member, self.c_name)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# https://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
from pathlib import Path
import datetime
import sphinx
import packaging.version

from sphinx.util.docfields import Field
from sphinx.locale import _


def get_version(root_path):
    """
    Reads the polybar version from the version.txt at the root of the repo.
    """
    path = Path(root_path) / "version.txt"
    with open(path, "r") as f:
        for line in f.readlines():
            if not line.startswith("#"):
                # NB: we can't parse it yet since sphinx could import
                # pkg_resources later on and it could patch packaging.version
                return line

    raise RuntimeError("No version found in {}".format(path))


sphinx_version = packaging.version.parse(sphinx.__version__)

# -- Project information -----------------------------------------------------


project = 'Polybar User Manual'
copyright = '2016-{}, Michael Carlberg & contributors'.format(
    datetime.datetime.now().year
)
author = 'Polybar Team'

# is whether we are on readthedocs.io
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if on_rtd:
    # On readthedocs, cmake isn't run so the version string isn't available
    version = os.environ.get('READTHEDOCS_VERSION', None)
else:
    # The short X.Y version
    version = '@APP_VERSION@'

# The full version, including alpha/beta/rc tags
release = version

# Set path to documentation
if on_rtd:
    # On readthedocs conf.py is already in the doc folder
    doc_path = '.'
else:
    # In all other builds conf.py is configured with cmake and put into the
    # build folder.
    doc_path = '@doc_path@'

# The version from the version.txt file. Since we are not always first
# configured by cmake, we don't necessarily have access to the current version
# number
version_txt = get_version(Path(doc_path).absolute().parent)

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.extlinks",
]

if on_rtd:
    extensions += [
        # The custom 404 page is only needed
        "notfound.extension",
        # The search extension works only on readthedocs
        # See https://readthedocs-sphinx-search.readthedocs.io
        "sphinx_search.extension",
    ]

# Add any paths that contain templates here, relative to this directory.
templates_path = [doc_path + '/_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
# language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

highlight_language = 'none'

smartquotes = False

# Quickly link to issues or PRs using :issue:`...` or :pull:`...`
if sphinx_version >= packaging.version.parse("4.0.0"):
    extlinks = {
        "issue": ("https://github.com/polybar/polybar/issues/%s", "#%s"),
        "pull": ("https://github.com/polybar/polybar/pull/%s", "PR #%s"),
    }
else:
    # Versions before 4.0 (e.g. on readthedocs) do not support %s in the
    # caption and simply append the value
    extlinks = {
        "issue": ("https://github.com/polybar/polybar/issues/%s", "#"),
        "pull": ("https://github.com/polybar/polybar/pull/%s", "PR #"),
    }

extlinks_detect_hardcoded_links = True


# -- Options for HTML output -------------------------------------------------

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {}

html_context = {
        'polybar_is_latest': version == 'latest',
}

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
if on_rtd or os.environ.get('USE_RTD_THEME', '0') == '1':
    html_theme = 'sphinx_rtd_theme'
    html_theme_options['collapse_navigation'] = False
    html_theme_options['style_external_links'] = True
else:
    html_theme = 'alabaster'
    html_theme_options['description'] = version

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = [doc_path + '/_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'polybardoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'polybar.tex', 'polybar Documentation',
     'Polybar Team', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        'man/polybar.1', 'polybar',
        'A fast and easy-to-use tool status bar', [], 1
    ),
    (
        'man/polybar-msg.1', 'polybar-msg',
        'Send IPC messages to polybar', [], 1
    ),
    (
        'man/polybar.5', 'polybar',
        'configuration file for polybar(1)', [], 5
    )
]

man_make_section_directory = False


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'polybar', 'polybar Documentation',
     author, 'polybar', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

# The 'versionadded' and 'versionchanged' directives are overridden.
suppress_warnings = ['app.add_directive']


def setup(app):

    # Adds a new directive for document a polybar config setting
    # Inside goes the description of the option as well as custom roles to
    # document the type, default value, etc:
    # .. poly-setting:: NAME
    #
    #   Description
    #   :type: ...
    #   :default: ...
    app.add_object_type(
        'poly-setting',
        'poly-setting',
        objname='configuration value',
        indextemplate='pair: %s; configuration value',
        doc_field_types=[
            Field('type',
                  label=_("Type"),
                  names=['type'],
                  has_arg=False,
                  ),
            Field('tags',
                  label=_("Available Tags"),
                  names=['tags'],
                  has_arg=False,
                  ),
            Field('tokens',
                  label=_("Supported Tokens"),
                  names=['tokens'],
                  has_arg=False,
                  ),
            Field('default',
                  label=_("Default Value"),
                  names=['default'],
                  has_arg=False,
                  ),
        ]
    )

    try:
        inject_version_directives(app)
    except NameError:
        # Function was not defined because sphinx version was too low
        pass


# It is not exactly clear in which version the VersionChange class was
# introduced, but we know it is available in at least 1.8.5.
# This feature is mainly needed for the online docs on readthedocs for the docs
# built from master, documentation built for proper releases should not even
# mention unreleased changes. Because of that it's not that important that this
# is added to local builds.
if sphinx_version >= packaging.version.parse("1.8.5"):

    from typing import List
    from docutils.nodes import Node
    from sphinx.domains.changeset import VersionChange

    def inject_version_directives(app):
        app.add_directive('deprecated', VersionDirective)
        app.add_directive('versionadded', VersionDirective)
        app.add_directive('versionchanged', VersionDirective)

    class VersionDirective(VersionChange):
        """
        Overwrites the Sphinx directive for versionchanged, versionadded, and
        deprecated and adds an unreleased tag to versions that are not yet
        released
        """

        def run(self) -> List[Node]:
            directive_version = packaging.version.parse(self.arguments[0])
            parsed_version_txt = packaging.version.parse(version_txt)

            if directive_version > parsed_version_txt:
                self.arguments[0] += " (unreleased)"

            return super().run()
